/*
Jmeno: Vojtech Esner
Datum: 17. cervence 2010
*/

#include <cstdio>
#include <vector>
#include <string>
#include "png.h"

using std::vector;
using std::string;

#define ERR 1
#define OK 0

// datová struktura pro uchování dat obrázku
struct ImagePNG {
    int width, height;
    int bitdepth;
    int colortype;
    unsigned char* pixdata;
};

// funkce pro načtení obrázku z PNG souboru
int loadPNG(const char* filename, ImagePNG& image) {
    // otevreni souboru
    FILE* fin = fopen(filename, "rb");
    if (fin == NULL) {
        fprintf(stderr, "soubor '%s' nelze otevrit\n", filename);
        return ERR;
    }

    // vytvoreni struktury pro nacteni png
    png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (png_ptr == NULL) {
        fclose(fin);
        return ERR;
    }

    // vytvoreni struktury pro informace o obrazku
    png_infop info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL) {
        fclose(fin);
        png_destroy_read_struct(&png_ptr, NULL, NULL);
        return ERR;
    }

    // sem se skoci v pripade chyby pri cteni png
    if (setjmp(png_jmpbuf(png_ptr))) {
        png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
        fclose(fin);
        return ERR;
    }

    // inicializace pro cteni ze souboru
    png_init_io(png_ptr, fin);

    // nacteni informaci o obrazku
    png_read_info(png_ptr, info_ptr);
    png_uint_32 width, height;
    int bit_depth, color_type, interlace_type;
    png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
        &interlace_type, NULL, NULL);

    // prevedeni z 16 bitu na kanal do 8 bitu na kanal (toto se tyka jen HDR obrazku, ktere jsou pomerne vzacne)
    if (bit_depth == 16) {
        png_set_strip_16(png_ptr);
        bit_depth = 8;
    }

    // pokud je v jednom bajtu vice pixelu, tyto se rozdeli kazdy do sveho bajtu
    png_set_packing(png_ptr);

    // aktualizace informaci o obrazku
    png_read_update_info(png_ptr, info_ptr);
    /*png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
        &interlace_type, NULL, NULL);*/

    // alokace pole ukazatelu na radky
    png_bytep* row_pointers = new png_bytep[height];

    // alokace pameti pro nacteni obrazku (ulozi se do jednorozmerneho pole)
    image.pixdata = new unsigned char[png_get_rowbytes(png_ptr, info_ptr)*height];

    // inicializace ostatnich clenu image
    image.width = width;
    image.height = height;
    image.bitdepth = bit_depth;
    image.colortype = color_type;

    // inicializace pole ukazatelu na radky
    for (png_uint_32 i = 0; i != height; ++i)
        row_pointers[i] = image.pixdata+png_get_rowbytes(png_ptr, info_ptr)*i;

    // nacteni obrazku
    png_read_image(png_ptr, row_pointers);

    // dokonceni cteni
    png_read_end(png_ptr, info_ptr);

    // uvolneni pameti a uzavreni souboru
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
    fclose(fin);
    delete[] row_pointers;

    return OK;
}

// zapise obrazek do tabulky (urceno pro paletovy nebo sedotonovy obrazek)
void writeTablePalDec(FILE* fout, const ImagePNG& image, int precision) {
    for (int y = 0; y != image.height; ++y) {
        for (int x = 0; x != image.width; ++x) {
            unsigned int pixel = image.pixdata[x+y*image.width];
            fprintf(fout, "%.*u ", precision, pixel);
        }
        fprintf(fout, "\n");
    }
}

// zapise obrazek do tabulky s pouzitim hexadecimalnich hodnot (urceno pro GA, RGB a RGBA obrazky)
void writeTableHex(FILE* fout, const ImagePNG& image, int channels) {
    for (int y = 0; y != image.height; ++y) {
        for (int x = 0; x != image.width; ++x) {
            for (int c = 0; c != channels; ++c) {
                unsigned int channel = image.pixdata[(x+y*image.width)*channels+c];
                fprintf(fout, "%.2x", channel);
            }
            fprintf(fout, " ");
        }
        fprintf(fout, "\n");
    }
}

// ulozi pixely obrazku do tabulky do textoveho souboru
int saveTXT(const char* filename, const ImagePNG& image) {
    /* mozne typy obrazku:
    PNG_COLOR_TYPE_GRAY (bit depths 1, 2, 4, 8, 16)
    PNG_COLOR_TYPE_GRAY_ALPHA (bit depths 8, 16)
    PNG_COLOR_TYPE_PALETTE (bit depths 1, 2, 4, 8)
    PNG_COLOR_TYPE_RGB (bit_depths 8, 16)
    PNG_COLOR_TYPE_RGB_ALPHA (bit_depths 8, 16)
    */
    FILE* fout = fopen(filename, "w");
    if (image.colortype == PNG_COLOR_TYPE_GRAY ||
        image.colortype == PNG_COLOR_TYPE_PALETTE) {
        if (image.bitdepth <= 2) {
            writeTablePalDec(fout, image, 1);
        }
        else if (image.bitdepth == 4) {
            writeTablePalDec(fout, image, 2);
        }
        else if (image.bitdepth == 8) {
            writeTablePalDec(fout, image, 3);
        }
        else {
            fprintf(stderr, "neznamy format obrazku\n");
            fclose(fout);
            return ERR;
        }
    }
    else if (image.colortype == PNG_COLOR_TYPE_GRAY_ALPHA) {
        writeTableHex(fout, image, 2);
    }
    else if (image.colortype == PNG_COLOR_TYPE_RGB) {
        writeTableHex(fout, image, 3);
    }
    else if (image.colortype == PNG_COLOR_TYPE_RGB_ALPHA) {
        writeTableHex(fout, image, 4);
    }
    else {
        fprintf(stderr, "neznamy format obrazku\n");
        fclose(fout);
        return ERR;
    }
    fclose(fout);
    return OK;
}

int main(int argc, char** argv) {
    if (argc == 2 || argc == 3) {
        string filename_png = argv[1];
        string filename_txt = argc == 3 ? argv[2] : "vystup.txt";
        ImagePNG image;
        if (loadPNG(filename_png.c_str(), image) == ERR) {
            fprintf(stderr, "chyba pri cteni obrazku\n");
            return 1;
        }
        if (saveTXT(filename_txt.c_str(), image) == ERR) {
            fprintf(stderr, "chyba pri zapisu tabulky\n");
            return 1;
        }
        return 0;
    } else {
        printf("Pouziti: pngtotxt nazev-vstupniho-souboru.png [nazev-vystupniho-souboru.txt]\n"
            "Pokud neni specifikovan nazev vystupniho souboru, pouzije se 'vystup.txt'.\n");
        return 1;
    }
    return 0;
}
